# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

"""Microsoft Azure Blob Service North plugin"""

import io
import asyncio
import copy
import csv
import dicttoxml
from datetime import datetime
import json
import numpy as np
from PIL import Image
import re

# HACK: FOGL-7781 This is a workaround to ensure Python executable path is correct 
import sys
sys.executable="/usr/bin/python3"
# 

from azure.storage.blob import ContainerClient, BlobClient

from foglamp.common import logger
from foglamp.plugins.north.common.common import *


__author__ = "Devki Nandan Ghildiyal"
__copyright__ = "Copyright (c) 2023 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__)

_DEFAULT_CONFIG = {
    "plugin": {
        "description": "Azure Blob North Plugin",
        "type": "string",
        "default": "azure-blob",
        "readonly": "true"
    },
    'accountName': {
        "description": "Azure Storage account name to which the Containers/Blobs will be published",
        "type": "string",
        "default":  "fog1storage1",
        "displayName": "Storage Account",
        "order": "1"
    },
    'accountKey': {
        "description": 'Azure storage account access key',
        "type": "password",
        "default": "foglamp",
        "displayName": "Account Key",
        "order": "2"
    },
    "imageFormat": {
        "description": "Image format to be saved into Azure blob",
        "type": "enumeration",
        "default": "PNG",
        "options": ["JPEG", "PNG"],
        "displayName": "Image Format",
        "order": "3",
        "group" : "Storage Format"
        
    },
    "dataFormat": {
        "description": "Data format to be saved into Azure blob",
        "type": "enumeration",
        "default": "JSON",
        "options": ["CSV", "JSON", "XML"],
        "displayName": "Data Format",
        "order": "4",
        "group" : "Storage Format"
    },
    "containerType": {
        "description": "Container creation preference",
        "type": "enumeration",
        "default": "Per Asset",
        "options": ["Per Asset", "Single Static"],
        "displayName": "Container Type",
        "order": "5",
        "group" : "Container Details"
    },
    "containerName": {
        "description": "Name of static container",
        "type": "string",
        "default": "",
        "displayName": "Static Container Name",
        "validity" : "containerType == \"Single Static\"",
        "order": "6",
        "group" : "Container Details"
    },
    "source": {
        "description": "The source of the data to send",
        "type": "enumeration",
        "default": "readings",
        "options": ["readings", "statistics"],
        "displayName": "Data Source"
    }
}


def plugin_info():
    return {
        'name': 'azure-blob',
        'version': '2.3.0',
        'type': 'north',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Registers azure-blob plugin handler

    Args:
        config: JSON configuration document for the North plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    handle = copy.deepcopy(config)
    handle['azure_blob_north'] = AzureBlobService(config)
    return handle


async def plugin_send(data, readings_block, stream_id):
    """ Sends the readings block data

        Args:
            data : Data to be sent to Azure Storage Container as Blobs
            readings_block :  Block of readings
            stream_id : Stream ID
    """
    try:
        azure_blob_north = data['azure_blob_north']
        is_data_sent, new_last_object_id, num_sent =  azure_blob_north._send(readings_block)
    except asyncio.CancelledError:
        pass
    except Exception as ex:
        _LOGGER.exception(ex, "Failed during send operation!")
    else:
        return is_data_sent, new_last_object_id, num_sent


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

        Args:
            handle : Plugin handle
            new_config : Changed configuration
    """
    diff = utils.get_diff(handle, new_config)
    _LOGGER.info("Config changed for azure-blob North plugin: [%s]", ', '.join(diff))


def plugin_shutdown(handle):
    """ Shut down the plugin

        Args:
            handle : Plugin handle
    """
    azure_blob_north = handle['azure_blob_north']
    azure_blob_north._shutdown()
    handle['azure_blob_north'] = None
    _LOGGER.debug("azure-blob north plugin shut down.")


class AzureBlobService(object):
    """Azure Blob Service Class"""
    
    __slots__ = [ "event_loop", "image_format", "data_format", "container_type", "static_container_name", "connection_string"]
    def __init__(self, config):

        self.event_loop = asyncio.get_event_loop()
        self.image_format = config['imageFormat']['value']
        self.data_format = config['dataFormat']['value']

        self.container_type = config['containerType']['value']
        self.static_container_name = config['containerName']['value']

        account_name = config['accountName']['value']
        account_key = config['accountKey']['value']
        if len(account_name.strip()) == 0 or len(account_key.strip()) == 0:
             _LOGGER.error("Invalid Storage Account name or key")

        conn_str_template = "DefaultEndpointsProtocol=https;AccountName={};AccountKey={};EndpointSuffix=core.windows.net"

        self.connection_string = conn_str_template.format(account_name, account_key)

    def _create_container_if_not_exists(self, _container_name):
        """ Create a container if it does not exist

        Args:
            _container_name : Name of the container to be created
        
        Returns:
            A tuple of status  of container if exists or not and response if container does not exist 
        """
       
        _container_client = ContainerClient.from_connection_string(conn_str=self.connection_string, container_name=_container_name)
        
        if not _container_client.exists():
            response =_container_client.create_container()
            _LOGGER.debug(response)
            return False, response
        else:
            _LOGGER.debug("Container {} already exists.".format(_container_name))
            return True, None
            

    def _sanitize(self, name, is_container_name):
        """ 
        Removes invalid characters from name 

        Args:
            name : name to be sanitized
            is_container_name :  True if conatiner name to be sanitized
        
        Returns:
            sanitized name
        """
         
        s_name = name.lower()
        if is_container_name:
            s_name = "".join(l for l in s_name if l in re.findall(r"[a-zA-Z0-9-]+", l))
        else:
            s_name = "".join(l for l in s_name if l in re.findall(r"[a-zA-Z0-9._-]+", l))
        _LOGGER.info("{} sanitized to {}".format(name, s_name))
        return s_name

    def _prepare_blob_name(self, is_image, asset_name, dp_name, ts):
        """ Create a blob name

        Args:
            is_image : True for image data
            asset_name : Asset name
            dp_name : Datapoint name
            ts : timestamp
        
        Returns:
            A tuple of blob name and extension
        """
        if is_image:
            blob_extension  = self.image_format
            blob_name = "{}-{}_{}.{}".format(ts, asset_name, dp_name, blob_extension)
        else:
            blob_extension  = self.data_format
            blob_name = "{}_{}_reading.{}".format(ts, asset_name, blob_extension)

        _LOGGER.debug("Blob Name: {} Extension: {}".format(blob_name, blob_extension))
        return blob_name, blob_extension

    def _prepare_blob_name_with_hints(self, is_image, asset_name, dp_name, blob_hints, ts):
        """ Create a blob name with hints

        Args:
            is_image : True for image data
            asset_name :  Asset name
            dp_name : Datapoint name
            blob_hints : Hints for Azure blob north plugin
            ts : timestamp

        Returns:
            A tuple of blob name and extension 
        """
        blob_name = "{}_{}_{}_{}".format(blob_hints['prefix'], ts, asset_name, blob_hints['postfix'])
        if is_image:
            blob_extension = blob_hints['imageformat']
            blob_name = "{}_{}.{}".format(blob_name, dp_name, blob_extension)
        else:
            blob_extension = blob_hints['dataformat']
            blob_name = "{}_reading.{}".format(blob_name, blob_extension)

        _LOGGER.debug("Blob Name: {} Extension: {}".format(blob_name, blob_extension))
        return blob_name, blob_extension
    
    def _create_image_blob(self, blob_service_client, image_data, image_type, blob_name, container_name):
        """ Create a image blob into Azure blob

        Args:
            blob_service_client :  Blob Service Client object
            image_data : Data for image
            image_type : Image type [JPEG/PNG]
            blob_name : Name of the blob to be saved
            container_name : Name of the container
        """
        image_data = Image.fromarray(image_data)
        image_bytes = io.BytesIO()
        image_data.save(image_bytes, format=image_type)
        image_bytes.seek(0)
        blob_service_client.upload_blob(image_bytes.read(), overwrite=True)
        _LOGGER.info("Image Blob [ {} ] published in container {}.".format(blob_name, container_name))

    def _create_non_image_blob(self, blob_service_client, reading_dict, data_type, blob_name, container_name):
        """ Create a non image blob into Azure blob

        Args:
            blob_service_client : Blob Service Client object
            reading_dict : Dictionary of all non image data
            blob_name : Name of the blob to be saved
            container_name : Name of the container
        """
        _dtype = data_type.lower()
        if _dtype == "json":
            # Send JSON
            jdoc = json.dumps(reading_dict)
            blob_service_client.upload_blob(jdoc, overwrite=True)
            _LOGGER.debug("JSON Blob [ {} ]  published in container {}".format(blob_name,  container_name))
        elif _dtype == "xml":
            # Send XML
            xml = dicttoxml.dicttoxml(reading_dict)
            blob_service_client.upload_blob(xml, overwrite=True)
            _LOGGER.debug("XML Blob [ {} ]  published in container {}".format(blob_name,  container_name))
        elif _dtype == "csv":
            # Send CSV 
            row_data = [reading_dict]
            field_names= reading_dict.keys()
            output = io.StringIO()
            writer = csv.DictWriter(output, fieldnames=field_names)
            writer.writeheader()
            writer.writerows(row_data)
            output.seek(0)
            blob_service_client.upload_blob(output.read(), overwrite=True)
            _LOGGER.debug("CSV Blob [ {} ]  published in container {}".format(blob_name,  container_name))
        else:
            _LOGGER.warning("Unsupported type {}".format(data_type))


    def _send(self, readings_block):
        """ Send readings block

        Args:
            readings_block : Block of readings
        """

        is_data_sent = False
        last_object_id = 0
        num_sent = 0

        _container_name = None
        blob_name, blob_ext = None,None

        def _get_reading_ts(reading_timestamp):
            fmt = "%Y-%m-%d %H:%M:%S.%f%z"
            if ":" == reading_timestamp[-3]:
                reading_timestamp = reading_timestamp[:-3]+reading_timestamp[-2:]
            return datetime.strptime(reading_timestamp, fmt).strftime("%Y_%m_%d_%H_%M_%S_%f")

        try:
            r_sent = False
            # Iterate through readings block
            for r in readings_block:
                r_sent = False
                _LOGGER.debug("Reading data [ {} ] ".format(r))
                last_object_id = r["id"]
               
                blob_service_client = None
                
                # Get blobhint
                blob_hints = r['reading'].get('blobhint')

                asset_name = self._sanitize(r['asset_code'], True)
                if self.container_type == "Per Asset":
                    _container_name = asset_name
                else:
                    _container_name = self._sanitize(self.static_container_name, True)

                try:
                    is_container_exists, response = self._create_container_if_not_exists(_container_name)
                    if not is_container_exists:
                        _LOGGER.debug("Container created. Response: {}".format(response))
                except:
                     _LOGGER.exception(ex, "Failed to check/create the container {}". format(_container_name))
                else:
                    # Iterate through all the reading object keys (datapoints)
                    reading_non_image_data = dict()   
                    ts = _get_reading_ts(r['user_ts'])
                    
                    for key in r['reading'].keys():
                        if key == "blobhint":
                            continue
                        _LOGGER.debug("Datapoint Name [ {} ] ".format(key))
                        data = r["reading"].get(key)
                        is_img = isinstance(data, np.ndarray) and (data).ndim > 1
                        
                        _LOGGER.debug("Reading [ {} ] ".format(data))
                    
                        datapoint_name = self._sanitize(key, False)
                    
                        if is_img:
                            if blob_hints:
                                blob_name, blob_ext = self._prepare_blob_name_with_hints(True, asset_name, datapoint_name, blob_hints, ts)
                            else:
                                blob_name, blob_ext = self._prepare_blob_name(True, asset_name, datapoint_name, ts)
                            blob_service_client = BlobClient.from_connection_string(conn_str=self.connection_string, container_name=_container_name, blob_name=blob_name)
                            self._create_image_blob(blob_service_client, data, blob_ext, blob_name, _container_name)
                            r_sent = True
                        else:
                            reading_non_image_data[key] = data

                        # iteration through reading object completed
                
                    if len(reading_non_image_data):
                        if blob_hints:
                            blob_name, blob_ext = self._prepare_blob_name_with_hints(False, asset_name, "", blob_hints, ts)
                        else:
                            blob_name, blob_ext = self._prepare_blob_name(False, asset_name, "", ts)
                        blob_service_client = BlobClient.from_connection_string(conn_str=self.connection_string, container_name=_container_name, blob_name=blob_name)
                        self._create_non_image_blob(blob_service_client, reading_non_image_data ,blob_ext, blob_name, _container_name)
                        r_sent = True
                    
                    if r_sent: 
                        num_sent += 1
                        is_data_sent = True

                # reading object of readings block
        except Exception as ex:
            _LOGGER.exception(ex, "Failed to publish the blob {} for {}". format(blob_name, _container_name))

        return is_data_sent, last_object_id, num_sent

    def _shutdown(self):
        """ Shutdown plugin """

        self.event_loop = None
